﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Reflection.Emit;
using UnityEngine;
using Verse;
using HarmonyLib;

namespace SmartTurretCovering
{
    static class HarmonyPatches
    {
        static readonly MethodInfo get_CellInfo = AccessTools.Property(typeof(LocalTargetInfo), nameof(LocalTargetInfo.Cell)).GetGetMethod();
        static readonly MethodInfo GetCoveredCellInfo = AccessTools.Method(typeof(HarmonyPatches), nameof(HarmonyPatches.GetCoveredCell));

        [HarmonyPatch(typeof(CoverUtility), nameof(CoverUtility.CalculateCoverGiverSet))]
        static class CalculateCoverGiverSet_Patch
        {
            [HarmonyTranspiler]
            static IEnumerable<CodeInstruction> Transpiler(IEnumerable<CodeInstruction> instructions)
            {
                var done = false;
                foreach (var instruction in instructions)
                {
                    if (!done && instruction.Calls(get_CellInfo))
                    {
                        yield return new CodeInstruction(OpCodes.Pop);
                        yield return new CodeInstruction(OpCodes.Ldarg_0);                           //target
                        yield return new CodeInstruction(OpCodes.Ldarg_1);                           //shooterLoc
                        yield return new CodeInstruction(OpCodes.Call, GetCoveredCellInfo);
                        done = true;
                    }
                    else
                    {
                        yield return instruction;
                    }
                }
            }
        }

        [HarmonyPatch(typeof(CoverUtility), nameof(CoverUtility.CalculateOverallBlockChance))]
        static class CalculateOverallBlockChance_Patch
        {
            [HarmonyTranspiler]
            static IEnumerable<CodeInstruction> Transpiler(IEnumerable<CodeInstruction> instructions)
            {
                var done = false;
                foreach (var instruction in instructions)
                {
                    if (!done && instruction.Calls(get_CellInfo))
                    {
                        yield return new CodeInstruction(OpCodes.Pop);
                        yield return new CodeInstruction(OpCodes.Ldarg_0);                           //target
                        yield return new CodeInstruction(OpCodes.Ldarg_1);                           //shooterLoc
                        yield return new CodeInstruction(OpCodes.Call, GetCoveredCellInfo);
                        done = true;
                    }
                    else
                    {
                        yield return instruction;
                    }
                }
            }
        }

        [HarmonyPatch(typeof(CoverUtility), "TryFindAdjustedCoverInCell")]
        static class TryFindAdjustedCoverInCell_Patch
        {
            [HarmonyTranspiler]
            static IEnumerable<CodeInstruction> Transpiler(IEnumerable<CodeInstruction> instructions)
            {
                var done = false;
                foreach (var instruction in instructions)
                {
                    if (!done && instruction.Calls(get_CellInfo))
                    {
                        yield return new CodeInstruction(OpCodes.Pop);
                        yield return new CodeInstruction(OpCodes.Ldarg_1);                           //target
                        yield return new CodeInstruction(OpCodes.Ldarg_0);                           //shooterLoc
                        yield return new CodeInstruction(OpCodes.Call, GetCoveredCellInfo);
                        done = true;
                    }
                    else
                    {
                        yield return instruction;
                    }
                }
            }

            [HarmonyPostfix]
            static void Postfix(bool __result, LocalTargetInfo target, IntVec3 adjCell, Map map, ref CoverInfo result)
            {
                if (__result)
                {
                    if (target.HasThing && target.Thing.def.category == ThingCategory.Building)
                    {
                        MethodInfo ShouldShowShotReportInfo = AccessTools.Method(typeof(TooltipGiverList), "ShouldShowShotReport");

                        if ((bool)ShouldShowShotReportInfo.Invoke(new TooltipGiverList() { }, new object[] { target.Thing }))
                        {
                            // target is a turret.
                            Thing cover = adjCell.GetCover(map);
                            result = new CoverInfo(cover, result.BlockChance / (float)Settings.CurrentFactor);
                        }
                    }
                }
            }
        }

        static IntVec3 GetCoveredCell(LocalTargetInfo target, IntVec3 shooterLoc)
        {
            if (target.HasThing)
            {
                float MinDistance = float.MaxValue;
                IntVec3 TargetLoc = IntVec3.Invalid;

                foreach (var ThingCell in target.Thing.OccupiedRect())
                {
                    float distance = ThingCell.DistanceTo(shooterLoc);
                    if (MinDistance > distance)
                    {
                        MinDistance = distance;
                        TargetLoc = ThingCell;
                    }
                }

                if (TargetLoc != IntVec3.Invalid)
                {
                    return TargetLoc;
                }
            }

            return target.Cell;
        }
    }
}
