﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Reflection.Emit;
using UnityEngine;
using Verse;
using HarmonyLib;

namespace SmartTurretCovering
{
    [HarmonyPatch(typeof(Verb_LaunchProjectile), "TryCastShot")]
    static class TryCastShot_Patch
    {
        [HarmonyTranspiler]
        static IEnumerable<CodeInstruction> Transpiler(IEnumerable<CodeInstruction> instructions, ILGenerator il)
        {
            MethodInfo IsTurretInfo = AccessTools.Method(typeof(TryCastShot_Patch), nameof(IsTurret));

            FieldInfo ThingCategoryInfo = AccessTools.Field(typeof(ThingDef), nameof(ThingDef.category));

            Label Label1 = il.DefineLabel();
            Label IL_035a = default;

            int InsertIndex = -1;

            var codes = new List<CodeInstruction>(instructions);
            for (var index = 0; index < codes.Count - 3; index++) // -3 since we will be checking i + 3
            {
                if (codes[index].LoadsField(ThingCategoryInfo))
                {
                    if (codes[index + 1].opcode == OpCodes.Ldc_I4_1)
                    {
                        if (codes[index + 2].opcode == OpCodes.Bne_Un_S)
                        {
                            IL_035a = (Label)codes[index + 2].operand;
                            codes[index + 2] = new CodeInstruction(OpCodes.Beq_S, Label1);

                            InsertIndex = index + 3;
                            codes[index + 3].labels.Add(Label1);
                            break;
                        }
                    }
                }
            }

            var InstructionsToInsert = new List<CodeInstruction>
            {
                new CodeInstruction(OpCodes.Ldarg_0),
                new CodeInstruction(OpCodes.Call, IsTurretInfo),
                new CodeInstruction(OpCodes.Brfalse_S, IL_035a)
            };

            if (InsertIndex != -1)
            {
                codes.InsertRange(InsertIndex, InstructionsToInsert);
            }

            return codes.AsEnumerable();
        }

        private static bool IsTurret(Verb_LaunchProjectile instance)
        {
            LocalTargetInfo target = instance.CurrentTarget;
            if (target.HasThing && target.Thing.def.category == ThingCategory.Building)
            {
                MethodInfo ShouldShowShotReportInfo = AccessTools.Method(typeof(TooltipGiverList), "ShouldShowShotReport");

                if ((bool)ShouldShowShotReportInfo.Invoke(new TooltipGiverList() { }, new object[] { target.Thing }))
                {
                    return true;
                }
            }

            return false;
        }
    }
}